import pytest
import numpy as np
from numpy.testing import assert_allclose
from ..formula import mean_from_string, _convert_token, _is_float, _token_to_mean
from ..formula import _tokenize_string, _parse_tokens, _eval_parsed_tokens
from ..MeanFunction import ConstantMean, Coefficient, LinearMean
from ..MeanFunction import MeanSum, MeanProduct, MeanPower, MeanComposite
try:
    import patsy
    no_patsy = False
    from ..formula import mean_from_patsy_formula, _term_to_mean
except ImportError:
    no_patsy = True

patsy_skip = pytest.mark.skipif(no_patsy, reason="patsy is needed to test formula parsing")

@patsy_skip
@pytest.mark.parametrize("code,inputdict,params,resulttype,result",
                         [("x[0] - 1",
                           {}     , np.ones(1)                , MeanProduct , np.array([1., 4.])),
                          ("a - 1",
                           {}     , np.ones(2)                , MeanProduct , np.array([1., 1.])),
                          ("a - 1",
                           {"a":1}, np.ones(1)                , MeanProduct , np.array([2., 5.])),
                          ("x[0]",
                           {}     , np.array([1., 2.])        , MeanSum    , np.array([3., 9.])),
                          ("x[1] - 1",
                           {}     , np.array([2.])            , MeanProduct, np.array([4., 10.])),
                          ("I ( inputs[0]^2) - 1",
                           {}     , np.ones(1)                , MeanProduct, np.array([1., 16.])),
                          ("x[0]*x[1]",
                           {}     , np.array([1., 2., 3., 4.]), MeanSum    , np.array([17., 104.])),
                          (patsy.ModelDesc.from_formula("x[0] - 1"),
                           {}     , np.ones(1)                , MeanProduct , np.array([1., 4.])),
                          (patsy.ModelDesc.from_formula("a - 1"),
                           {}     , np.ones(2)                , MeanProduct , np.array([1., 1.])),
                          (patsy.ModelDesc.from_formula("a - 1"),
                           {"a":1}, np.ones(1)                , MeanProduct , np.array([2., 5.])),
                          (patsy.ModelDesc.from_formula("x[0]"),
                           {}     , np.array([1., 2.])        , MeanSum    , np.array([3., 9.])),
                          (patsy.ModelDesc.from_formula("x[1] - 1"),
                           {}     , np.array([2.])            , MeanProduct, np.array([4., 10.])),
                          (patsy.ModelDesc.from_formula("I ( inputs[0]^2) - 1"),
                           {}     , np.ones(1)                , MeanProduct, np.array([1., 16.])),
                          (patsy.ModelDesc.from_formula("x[0]*x[1]"),
                           {}     , np.array([1., 2., 3., 4.]), MeanSum    , np.array([17., 104.]))])
def test_mean_from_patsy_formula(code, inputdict, params, resulttype, result):
    "test the mean_from_patsy_formula function"

    x = np.array([[1., 2., 3.], [4., 5., 6.]])

    mf = mean_from_patsy_formula(code, inputdict)

    assert isinstance(mf, resulttype)
    assert mf.get_n_params(x) == len(params)

    assert_allclose(mf.mean_f(x, params), result)

@pytest.mark.parametrize("code,inputdict,params,resulttype,result",
                         [("x[0]",
                           {}     , np.zeros(0)       , LinearMean , np.array([1., 4.])),
                          ("y = x[0]",
                           {}     , np.zeros(0)       , LinearMean , np.array([1., 4.])),
                          ("y ~ x[0]",
                           {}     , np.zeros(0)       , LinearMean , np.array([1., 4.])),
                          ("~ x[0]",
                           {}     , np.zeros(0)       , LinearMean , np.array([1., 4.])),
                          ("= x[0]",
                           {}     , np.zeros(0)       , LinearMean , np.array([1., 4.])),
                          ("a",
                           {}     , np.ones(1)        , Coefficient, np.array([1., 1.])),
                          ("a",
                           {"a":1}, np.zeros(0)       , LinearMean , np.array([2., 5.])),
                          ("a + b*x[0]",
                           {}     , np.array([1., 2.]), MeanSum    , np.array([3., 9.])),
                          ("a*x[1]",
                           {}     , np.array([2.])    , MeanProduct, np.array([4., 10.])),
                          ("inputs[0]**2",
                           {}     , np.zeros(0)       , MeanPower  , np.array([1., 16.])),
                          ("I ( inputs[0]^2)",
                           {}     , np.zeros(0)       , MeanPower  , np.array([1., 16.])),
                          ("(a + b*x[0])(x[0]*x[1])",
                           {}     , np.array([1., 2.]), MeanComposite, np.array([5., 41.]))])
def test_mean_from_string(code, inputdict, params, resulttype, result):
    "test the mean_from_string function"

    x = np.array([[1., 2., 3.], [4., 5., 6.]])

    mf = mean_from_string(code, inputdict)

    assert isinstance(mf, resulttype)
    assert mf.get_n_params(x) == len(params)

    assert_allclose(mf.mean_f(x, params), result)

def test_mean_from_string_failures():
    "test that mean_from_string correctly fails"

    with pytest.raises(AssertionError):
        mean_from_string(1)

@patsy_skip
def test_term_to_mean():
    "test the _term_to_mean function"

    t = patsy.ModelDesc.from_formula("x[0]").rhs_termlist

    x = np.array([[1., 2., 3.], [4., 5., 6.]])

    params = np.array([2.])

    type_expected = [ Coefficient, MeanProduct ]
    mean_expected = [ np.array([2., 2.]), np.array([2., 8.]) ]

    for term, type_exp, mean_exp in zip(t, type_expected, mean_expected):
        mf = _term_to_mean(term)
        assert mf.get_n_params(x) == len(params)
        assert isinstance(mf, type_exp)
        assert_allclose(mf.mean_f(x, params), mean_exp)

@patsy_skip
def test_term_to_mean_failures():
    "test situations where term_to_mean should fail"

    with pytest.raises(AssertionError):
        _term_to_mean(1)

    t = patsy.ModelDesc.from_formula("x").rhs_termlist[0]
    t.factors = [1, 2]

    with pytest.raises(AssertionError):
        _term_to_mean(t)

@pytest.mark.parametrize("token,inputdict,result", [("x[1]", {}, "x[1]"),
                                                   ("inputs[0]", {}, "x[0]"),
                                                   ("a", {}, "a"),
                                                   ("a", {"a":0}, "x[0]"),
                                                   ("d", {"d":1}, "x[1]")])
def test_convert_token(token, inputdict, result):
    "test the function to convert tokens"

    assert _convert_token(token, inputdict) == result

def test_convert_token_failures():
    "test situations where _convert_token should fail"

    with pytest.raises(AssertionError):
        _convert_token(1)

def test_is_float():
    "test the _is_float function"

    assert _is_float("1")
    assert _is_float("-2.e-4")
    assert not _is_float("a")

@pytest.mark.parametrize("token,inputdict,params,resulttype,result",
                         [("2."  , {}     , np.zeros(0), ConstantMean, np.array([2., 2.])),
                          ("a"   , {}     , np.ones(1) , Coefficient , np.array([1., 1.])),
                          ("x[0]", {}     , np.zeros(0), LinearMean  , np.array([1., 4.])),
                          ("a"   , {"a":0}, np.zeros(0), LinearMean  , np.array([1., 4.])),
                          ("x[1]", {}     , np.zeros(0), LinearMean  , np.array([2., 5.]))])
def test_token_to_mean(token, inputdict, params, resulttype, result):
    "test the _token_to_mean function"

    x = np.array([[1., 2., 3.], [4., 5., 6.]])

    mf = _token_to_mean(token, inputdict)

    assert isinstance(mf, resulttype)
    assert mf.get_n_params(x) == len(params)

    assert_allclose(mf.mean_f(x, params), result)

def test_token_to_mean_failures():
    "test situation where token_to_mean should fail"

    with pytest.raises(AssertionError):
        _token_to_mean(1)

    with pytest.raises(ValueError):
        _token_to_mean("x[0")

    with pytest.raises(ValueError):
        _token_to_mean("x(1]")

    with pytest.raises(ValueError):
        _token_to_mean("x[a]")

    with pytest.raises(AssertionError):
        _token_to_mean("x[-19]")

@pytest.mark.parametrize("code,result", [("a", ["a"]),
                                         ("y = a", ["a"]),
                                         ("y ~ a", ["a"]),
                                         ("= a", ["a"]),
                                         ("~ a", ["a"]),
                                         ("a + b", ["a", "+", "b"]),
                                         ("a*b+ c", ["a", "*", "b", "+", "c"]),
                                         ("x[0]", ["x[0]"]),
                                         ("x[ 0 ]", ["x[0]"]),
                                         ("x**2", ["x", "^", "2"]),
                                         ("x^2", ["x", "^", "2"]),
                                         ("c(d)", ["c", "(", "d", ")"]),
                                         ("a*(b + c)**3 + x[a]", ["a", "*", "(", "b", "+", "c", ")",
                                                                  "^", "3", "+", "x[a]"])])
def test_tokenize_string(code, result):
    "test the _tokenize_string function"

    assert _tokenize_string(code) == result

def test_tokenize_string_failures():
    "test situations where _tokenize_string should fail"

    with pytest.raises(AssertionError):
        _tokenize_string(1)

    with pytest.raises(SyntaxError):
        _tokenize_string("x[")

    with pytest.raises(SyntaxError):
        _tokenize_string("x]")

    with pytest.raises(SyntaxError):
        _tokenize_string("x[1] + x]")

    with pytest.raises(SyntaxError):
        _tokenize_string("x[(0)]")

    with pytest.raises(SyntaxError):
        _tokenize_string("call")

    with pytest.raises(SyntaxError):
        _tokenize_string("a + b = c")

    with pytest.raises(SyntaxError):
        _tokenize_string("a + b ~ c")

@pytest.mark.parametrize("token_list,output_list",
                         [(["a"], ["a"]),
                          (["a", "+", "b"], ["a", "b", "+"]),
                          (["a", "*", "b", "+", "c"], ["a", "b", "*", "c", "+"]),
                          (["x", "^", "2"], ["x", "2", "^"]),
                          (["c", "(", "d", ")"], ["c", "d", "call"]),
                          (["a", "*", "(", "b", "+", "c", ")", "^", "3", "+", "x[0]"],
                           ["a", "b", "c", "+", "3", "^", "*", "x[0]", "+"])])
def test_parse_tokens(token_list, output_list):
    "test the parse_tokens function"

    assert _parse_tokens(token_list) == output_list

def test_parse_tokens_failures():
    "test situation where parse_tokens fails"

    with pytest.raises(AssertionError):
        _parse_tokens("a")

    with pytest.raises(AssertionError):
        _parse_tokens([1])

    with pytest.raises(SyntaxError):
        _parse_tokens(["x" "+", "(", "a", "+", "b"])

    with pytest.raises(SyntaxError):
        _parse_tokens(["x" "*", "a", "+", "b", ")"])

    with pytest.raises(SyntaxError):
        _parse_tokens(["="])

    with pytest.raises(SyntaxError):
        _parse_tokens(["~"])

@pytest.mark.parametrize("stack,inputdict,params,resulttype,result",
                         [(["x[0]"],
                           {}     , np.zeros(0)       , LinearMean , np.array([1., 4.])),
                          (["a"],
                           {}     , np.ones(1),        Coefficient , np.array([1., 1.])),
                          (["a"],
                           {"a":1}, np.zeros(0)       , LinearMean , np.array([2., 5.])),
                          (["a", "b", "x[0]", "*", "+"],
                           {}     , np.array([1., 2.]), MeanSum    , np.array([3., 9.])),
                          (["a", "x[1]", "*"],
                           {}     , np.array([2.])    , MeanProduct, np.array([4., 10.])),
                          (["inputs[0]", "2", "^"],
                           {}     , np.zeros(0)       , MeanPower  , np.array([1., 16.])),
                          (["I", "inputs[0]", "2", "^", "call"],
                           {}     , np.zeros(0)       , MeanPower  , np.array([1., 16.])),
                          (["a", "b", "x[0]", "*", "+", "x[0]", "x[1]", "*", "call"],
                           {}     , np.array([1., 2.]), MeanComposite, np.array([5., 41.]))])
def test_eval_parsed_tokens(stack, inputdict, params, resulttype, result):
    "test the eval_parsed_tokens function"

    x = np.array([[1., 2., 3.], [4., 5., 6.]])

    mf = _eval_parsed_tokens(stack, inputdict)

    assert isinstance(mf, resulttype)
    assert mf.get_n_params(x) == len(params)

    assert_allclose(mf.mean_f(x, params), result)

def test_eval_parsed_tokens_failures():
    "test situations where eval_parsed_tokens should fail"

    with pytest.raises(AssertionError):
        _eval_parsed_tokens(1)

    with pytest.raises(AssertionError):
        _eval_parsed_tokens([1])

    with pytest.raises(SyntaxError):
        _eval_parsed_tokens(["a", "+"])

    with pytest.raises(SyntaxError):
        _eval_parsed_tokens(["a", "I", "+"])

    with pytest.raises(SyntaxError):
        _eval_parsed_tokens(["I", "a", "+"])

    with pytest.raises(SyntaxError):
        _eval_parsed_tokens(["a", "b"])

    with pytest.raises(SyntaxError):
        _eval_parsed_tokens(["="])

    with pytest.raises(SyntaxError):
        _eval_parsed_tokens(["~"])



'''
This benchmark performs convergence tests on a single emulator with variable numbers of input parameters.
The example is based on the Rosenbrock function (see https://www.sfu.ca/~ssurjano/rosen.html). This
function can be defined in an artibrary number of dimensions, so it provides a useful test for how
emulators based on increasing numbers of parameters perform as the size of the training data is varied.
As the number of training points increases, the prediction error and prediction variance should
decrease. However, this will depend on the number of dimensions in the function -- in general, the
size of the input space grows exponentially with the number of dimensions, while the samples
drawn here grow linearly with the number of dimensions. Thus, the higher dimensional emulators will
perform worse for the same number of samples per dimension.
'''

import numpy as np
from mogp_emulator import GaussianProcess, fit_GP_MAP
from mogp_emulator import MonteCarloDesign, LatinHypercubeDesign
try:
    import matplotlib.pyplot as plt
    makeplots = True
except ImportError:
    makeplots = False

def rosenbrock(x):
    """
    Rosenbrock function in an arbitrary number of dimensions. Input x should be a 2D array of shape
    (n, D), where the first index represents the different places where the function is evaluated and
    the second index represents the dimensions of the function.
    see https://www.sfu.ca/~ssurjano/rosen.html for more information
    """
    x = np.array(x)
    if len(x.shape) == 1:
        x = np.reshape(x, (1, len(x)))
    assert len(x.shape) == 2
    return np.sum( 100. * ( x[:,1:] - x[:,:-1]**2 )**2 + ( x[:,:-1] - 1. )**2, axis = 1)


def generate_input_data(n_simulations, n_dimensions, method = "random"):
    "Generate n_simulations evaluation points the n_dimensions dimensional Rosenbrock function"

    n_simulations = int(n_simulations)
    n_dimensions = int(n_dimensions)
    assert(n_simulations > 0)
    assert(n_dimensions > 0)
    assert method == "random" or method == "lhd"

    if method == "random":
        ed = MonteCarloDesign(n_dimensions, (-5., 10.))
    elif method == "lhd":
        ed = LatinHypercubeDesign(n_dimensions, (-5., 10.))
    inputs = ed.sample(n_simulations)
    return inputs

def generate_training_data(n_simulations, n_dimensions):
    "Generate n_simulations input data for an n_dimensions dimensional Rosenbrock function"

    inputs = generate_input_data(n_simulations, n_dimensions, method = "lhd")
    targets = rosenbrock(inputs)

    return inputs, targets

def generate_test_data(n_testing, n_dimensions):
    "Generate n_testing points for testing the accuracy of an emulator"

    testing = generate_input_data(n_testing, n_dimensions, method = "random")
    test_targets = rosenbrock(testing)

    return testing, test_targets

def run_model(n_simulations, n_dimensions, n_testing):
    "Generate training data, fit emulator, and test model accuracy on random points, returning RMSE"

    inputs, targets = generate_training_data(n_simulations, n_dimensions)

    norm_const = np.mean(targets)

    gp = GaussianProcess(inputs, targets)
    gp = fit_GP_MAP(gp)

    testing, test_targets = generate_test_data(n_testing, n_dimensions)

    test_vals, unc, deriv = gp.predict(testing, deriv = False, unc = True)

    return (np.sqrt(np.sum((test_vals - test_targets)**2)/float(n_testing))/norm_const,
            np.sqrt(np.sum(unc**2)/float(n_testing))/norm_const**2)

def plot_model_errors(dimension_list, simulation_list, error, unc, n_testing):
    "Makes plot showing accuracy of emulator as a function of n_simulations"

    plt.figure(figsize=(4,3))
    for dim, err in zip(dimension_list, error):
        plt.semilogy(np.array(simulation_list), err, '-o', label = str(dim)+" dimensions")
    plt.legend()
    plt.xlabel('Number of simulations per dimension')
    plt.ylabel('Relative prediction RMSE')
    plt.title('Error for '+str(n_testing)+' predictions\nusing Rosenbrock function')
    plt.savefig('rosenbrock_error.png',bbox_inches='tight')

    plt.figure(figsize=(4,3))
    for dim, un in zip(dimension_list, unc):
        plt.semilogy(np.array(simulation_list), un, '-o', label = str(dim)+" dimensions")
    plt.legend()
    plt.xlabel('Number of simulations per dimension')
    plt.ylabel('Relative prediction variance')
    plt.title('Uncertainty for '+str(n_testing)+' predictions\nusing Rosenbrock function')
    plt.savefig('rosenbrock_unc.png',bbox_inches='tight')

def run_all_models(n_testing, dimension_list, simulation_list, n_iter = 10):
    "Runs all simulations, printing out results and optionally makes plots of results"

    n_simtrials = len(simulation_list)
    n_dimtrials = len(dimension_list)

    errors = np.zeros((n_dimtrials, n_simtrials, n_iter))
    uncs = np.zeros((n_dimtrials, n_simtrials, n_iter))

    for dim_index in range(n_dimtrials):
        for sim_index in range(n_simtrials):
            for iteration in range(n_iter):
                errors[dim_index, sim_index, iteration], uncs[dim_index, sim_index, iteration] = (
                                    run_model(simulation_list[sim_index]*dimension_list[dim_index], dimension_list[dim_index], n_testing))

    error = np.mean(errors, axis = -1)
    unc = np.mean(uncs, axis = -1)

    print("\n")
    print("Convergence test results:")
    print("Num. simulations/dim   Num. dimensions   Relative prediction RMSE")
    for dim, err_list in zip(dimension_list, error):
        for sim, err in zip(simulation_list, err_list):
            print('{:23}{:18}{}'.format(str(sim*dim), str(dim), str(err)))

    print("\n")
    print("Num. simulations/dim   Num. dimensions   Relative prediction variance")
    for dim, un_list in zip(dimension_list, unc):
        for sim, un in zip(simulation_list, un_list):
            print('{:23}{:18}{}'.format(str(sim*dim), str(dim), str(un)))

    if makeplots:
        plot_model_errors(dimension_list, simulation_list, error, unc, n_testing)

if __name__ == '__main__':
    run_all_models(100, [4, 6, 8], [i for i in range(2, 16)], 10)


import numpy as np
from numpy.testing import assert_allclose
import pytest
from ..HistoryMatching import HistoryMatching
from ..fitting import fit_GP_MAP
from ..GaussianProcess import GaussianProcess, PredictResult

def get_y_simulated_1D(x):
    n_points = len(x)
    f = np.zeros(n_points)
    for i in range(n_points):
        f[i] = np.sin(2.*np.pi*x[i] / 50.)
    return f

def test_sanity_checks():
    "test basic functioning of HistoryMatching"

    # Create a gaussian process
    x_training = np.array([
        [0.],
        [10.],
        [20.],
        [30.],
        [43.],
        [50.]
    ])

    y_training = get_y_simulated_1D(x_training)

    np.random.seed(47)
    gp = fit_GP_MAP(x_training, y_training)

    # Define observation and implausibility threshold
    obs = [-0.8, 0.0004]

    # Coords to predict
    n_rand = 2000
    x_predict_min = -3
    x_predict_max = 53
    x_predict = np.random.rand(n_rand)
    x_predict = np.sort(x_predict, axis=0)
    x_predict *= (x_predict_max - x_predict_min)
    x_predict += x_predict_min
    x_predict = x_predict[:, None]

    coords = x_predict

    expectations = gp.predict(coords)

    # Create History Matching Instance
    print("---TEST INPUTS---")
    print("No Args")
    hm = HistoryMatching()
    hm.status()

    print("Obs Only a - list")
    hm = HistoryMatching(obs=obs)
    hm.status()

    print("Obs only b - single-element list")
    hm = HistoryMatching(obs=[3.])
    hm.status()

    print("Obs only c - single-value")
    hm = HistoryMatching(obs=3.)
    hm.status()

    print("gp Only")
    hm = HistoryMatching(gp=gp)
    hm.status()

    print("Coords only a - 2d ndarray")
    hm = HistoryMatching(coords=coords)
    hm.status()

    print("Coords only b - 1d ndarray")
    hm = HistoryMatching(coords=np.random.rand(n_rand))
    hm.status()

    print("Coords only c - list")
    hm = HistoryMatching(coords=[a for a in range(n_rand)])
    hm.status()

    print("Expectation only")
    hm = HistoryMatching(expectations=expectations)
    hm.status()

    print("Threshold Only")
    hm = HistoryMatching(threshold=3.)
    hm.status()


    print("---TEST ASSIGNMENT---")
    print("Assign gp")
    hm = HistoryMatching(obs)
    hm.status()
    hm.set_gp(gp)
    hm.status()

    print("Assign Obs")
    hm = HistoryMatching(gp)
    hm.status()
    hm.set_obs(obs)
    hm.status()

    print("Assign Coords")
    hm = HistoryMatching()
    hm.status()
    hm.set_coords(coords)
    hm.status()

    print("Assign Expectations")
    hm = HistoryMatching()
    hm.status()
    hm.set_expectations(expectations)
    hm.status()

    print("Assign Threshold")
    hm = HistoryMatching()
    hm.status()
    hm.set_threshold(3.)
    hm.status()


    print("---TEST IMPLAUSABILIY---")
    print("implausibility test a - no vars")
    hm = HistoryMatching(obs=obs, gp=gp, coords=coords)
    I = hm.get_implausibility()

    print("implausibility test b - single value")
    hm = HistoryMatching(obs=obs, gp=gp, coords=coords)
    I = hm.get_implausibility(7.)

def test_HistoryMatching_init():
    "test the init method of HistoryMatching"

    # basic functionality

    hm = HistoryMatching()

    assert hm.gp == None
    assert hm.obs == None
    assert hm.coords == None
    assert hm.expectations == None
    assert hm.ndim == None
    assert hm.ncoords == None
    assert_allclose(hm.threshold, 3.)
    assert hm.I == None
    assert hm.NROY == None
    assert hm.RO == None

    # with gp/coords, obs, and threshold

    gp = GaussianProcess(np.reshape(np.linspace(0., 1.), (-1, 1)), np.linspace(0., 1.))
    coords = np.array([[0.2], [0.4]])
    hm = HistoryMatching(gp=gp, obs=1., coords=coords, expectations=None, threshold=5.)

    assert hm.gp == gp
    assert_allclose(hm.obs, [1., 0.])
    assert_allclose(hm.coords, coords)
    assert hm.expectations is None
    assert hm.ndim == 1
    assert hm.ncoords == len(coords)
    assert_allclose(hm.threshold, 5.)
    assert hm.I == None
    assert hm.NROY == None
    assert hm.RO == None

    # with obs, expectations, coords, and threshold

    expectations = PredictResult(mean=np.array([1.]), unc=np.array([0.2]), deriv=np.array([[0.1]]))
    hm = HistoryMatching(gp=None, obs=[1., 0.1], coords=None, expectations=expectations,
                         threshold=5.)

    assert hm.gp == None
    assert_allclose(hm.obs, [1., 0.1])
    assert hm.coords == None
    for a in range(3):
        assert_allclose(hm.expectations[a], expectations[a])
    assert hm.ndim == None
    assert hm.ncoords == len(expectations[0])
    assert_allclose(hm.threshold, 5.)
    assert hm.I == None
    assert hm.NROY == None
    assert hm.RO == None


def test_HistoryMatching_select_expectations():
    "test the select_expectations method of HistoryMatching"

    # correct functionality

    expectations = PredictResult(mean=np.array([2., 10.]), unc=np.array([0., 0.]),
                                 deriv=np.array([[1., 2.]]))
    hm = HistoryMatching(obs=[1., 1.], expectations=expectations)

    expectations_new = hm._select_expectations()

    for a in range(3):
        assert_allclose(expectations[a], expectations_new[a])

    np.random.seed(57483)
    gp = fit_GP_MAP(np.reshape(np.linspace(0., 1.), (-1, 1)), np.linspace(0., 1.))
    coords = np.array([[0.1], [1.]])
    obs = [1., 0.01]
    expectations = gp.predict(coords)

    hm = HistoryMatching(gp=gp, obs=obs, coords=coords)

    expectations_new = hm._select_expectations()

    for a, b in zip(expectations, expectations_new):
        assert_allclose(a, b)

    # ncoords somehow not set

    hm.ncoords = None
    with pytest.raises(ValueError):
        hm._select_expectations()

    # both coords and expectations set

    hm = HistoryMatching(gp=gp, obs=obs, coords=coords, expectations=expectations)

    with pytest.raises(ValueError):
        hm._select_expectations()

    # if no expectations provided, fails

    hm = HistoryMatching(obs=obs)

    with pytest.raises(ValueError):
        hm._select_expectations()

def test_HistoryMatching_get_implausibility():
    "test the get_implausibility method of HistoryMatching"

    # correct functionality

    expectations = PredictResult(mean=np.array([2., 10.]), unc=np.array([0., 0.]), deriv=np.array([[1., 2.]]))
    hm = HistoryMatching(obs=[1., 1.], expectations=expectations)
    I = hm.get_implausibility()

    assert_allclose(I, [1., 9.])
    assert_allclose(hm.I, [1., 9.])

    I = hm.get_implausibility(1.)

    assert_allclose(I, [1./np.sqrt(2.), 9./np.sqrt(2.)])
    assert_allclose(hm.I, [1./np.sqrt(2.), 9./np.sqrt(2.)])

    np.random.seed(57483)
    gp = fit_GP_MAP(np.reshape(np.linspace(0., 1.), (-1, 1)), np.linspace(0., 1.))
    coords = np.array([[0.1], [1.]])
    obs = [1., 0.01]
    mean, unc, _ = gp.predict(coords)
    I_exp = np.abs(mean - obs[0])/np.sqrt(unc + obs[1])

    hm = HistoryMatching(gp=gp, obs=obs, coords=coords)
    I = hm.get_implausibility()

    assert_allclose(I, I_exp)
    assert_allclose(hm.I, I_exp)

    # no observations

    hm = HistoryMatching(expectations=expectations)

    with pytest.raises(ValueError):
        hm.get_implausibility()

    # negative variance for model discrepancy

    hm = HistoryMatching(obs=[1., 1.], expectations=expectations)

    with pytest.raises(AssertionError):
        hm.get_implausibility(-1.)

def test_HistoryMatching_get_NROY():
    "test the get_NROY method of HistoryMatching"

    # correct functionality

    hm = HistoryMatching(obs=[1., 1.], expectations=PredictResult(mean=np.array([2., 10.]),
                                                                  unc=np.array([0., 0.]),
                                                                  deriv=np.array([[1., 2.]])))
    hm.get_implausibility()

    NROY = hm.get_NROY()

    assert NROY == [0]

    hm = HistoryMatching(obs=[1., 0.], expectations=PredictResult(mean=np.array([2., 10.]),
                                                                  unc=np.array([0., 0.]),
                                                                  deriv=np.array([[1., 2.]])))

    NROY = hm.get_NROY(1.)

    assert NROY == [0]

def test_HistoryMatching_get_RO():
    "test the get_RO method of HistoryMatching"

    # correct functionality

    hm = HistoryMatching(obs=[1., 1.], expectations=PredictResult(mean=np.array([2., 10.]),
                                                                  unc=np.array([0., 0.]),
                                                                  deriv=np.array([[1., 2.]])))
    hm.get_implausibility()

    RO = hm.get_RO()

    hm = HistoryMatching(obs=[1., 0.], expectations=PredictResult(mean=np.array([2., 10.]),
                                                                  unc=np.array([0., 0.]),
                                                                  deriv=np.array([[1., 2.]])))

    RO = hm.get_RO(1.)

    assert RO == [1]

def test_HistoryMatching_set_gp():
    'test the set_gp method of HistoryMatching'

    # correct functionality

    gp = GaussianProcess(np.reshape(np.linspace(0., 1.), (-1, 1)), np.linspace(0., 1.))

    hm = HistoryMatching()
    hm.set_gp(gp)

    assert hm.gp == gp

    # bad type for GP

    with pytest.raises(TypeError):
        hm.set_gp(gp=1.)

def test_HistoryMatching_set_obs():
    'test the set_obs method of HistoryMatching'

    # correct functionality

    obs = [1., 0.1]

    hm = HistoryMatching()
    hm.set_obs(obs)

    assert_allclose(hm.obs, obs)

    obs = 1.
    hm.set_obs(obs)

    assert_allclose(hm.obs, [obs, 0.])

    # wrong number of values for obs

    with pytest.raises(ValueError):
        hm.set_obs([1., 2., 3.])

    # negative variance

    with pytest.raises(AssertionError):
        hm.set_obs([1., -1.])

    # bad values for obs

    with pytest.raises(TypeError):
        hm.set_obs("abc")

def test_HistoryMatching_set_coords():
    "test the set_coords method of HistoryMatching"

    # correct functionality

    coords = np.array([[1.]])

    hm = HistoryMatching()
    hm.set_coords(coords)

    assert_allclose(hm.coords, coords)

    coords = np.array([1.])

    hm.set_coords(coords)

    assert_allclose(hm.coords, np.reshape(coords, (1, 1)))

    # confirm can reset coords

    hm.set_coords(None)

    assert hm.coords == None

def test_HistoryMatching_set_expectations():
    "test the set_expectations method of HistoryMatching"

    # correct functionality

    expectations = PredictResult(mean=np.array([0.]), unc=np.array([0.1]), deriv=np.array([[2.]]))

    hm = HistoryMatching()
    hm.set_expectations(expectations)

    for a in range(3):
        assert_allclose(hm.expectations[a], expectations[a])

    # confirm ability to reset expectations

    hm.set_expectations(None)
    assert hm.expectations == None

    # verify derivatives can be None

    expectations = PredictResult(mean=np.array([0.]), unc=np.array([0.1]), deriv=None)

    hm = HistoryMatching()
    hm.set_expectations(expectations)

    for a in range(2):
        assert_allclose(hm.expectations[a], expectations[a])
    assert hm.expectations[2] is None

    # negative variance

    expectations = PredictResult(mean=np.array([0.]), unc=np.array([-0.1]), deriv=np.array([[2.]]))

    with pytest.raises(AssertionError):
        hm.set_expectations(expectations)

    # bad values for expectations

    expectations = PredictResult(mean=np.array([0.1]), unc=1., deriv=1.)
    hm = HistoryMatching()

    with pytest.raises(TypeError):
        hm.set_expectations(expectations)

    expectations = PredictResult(mean=np.array([0.1]), unc=np.array([1., 2.]), deriv=1.)

    with pytest.raises(TypeError):
        hm.set_expectations(expectations)

    expectations = PredictResult(mean=np.array([1.]), unc=np.array([1., 2.]), deriv=None)

    with pytest.raises(ValueError):
        hm.set_expectations(expectations)


def test_HistoryMatching_set_threshold():
    "test the set_threshold method of HistoryMatching"

    # correct functionality

    hm = HistoryMatching()

    hm.set_threshold(1.)

    assert_allclose(hm.threshold, 1.)

    # threshold cannot be negative

    with pytest.raises(AssertionError):
        hm.set_threshold(-1.)

def test_HistoryMatching_update():
    "test the update method of HistoryMatching"

    # correct functionality

    hm = HistoryMatching()

    hm.coords = np.array([[1.]])

    hm.update()

    assert hm.ncoords == 1
    assert hm.ndim == 1

    hm = HistoryMatching()

    hm.expectations = PredictResult(mean=np.array([0.]), unc=np.array([0.1]), deriv=np.array([[2.]]))
    hm.update()

    assert hm.ncoords == 1

def test_HistoryMatching_str():
    "test the str method of HistoryMatching"

    # correct functionality

    hm = HistoryMatching()

    str_exp = ("History Matching tools created with:\n" +
               "Gaussian Process: None\n" +
               "Observations: None\n" +
               "Coords: None\n" +
               "Expectations: None\n" +
               "No. of Input Dimensions: None\n" +
               "No. of Descrete Expectation Values: None\n" +
               "I_threshold: {}\n" +
               "I: None\n" +
               "NROY: None\n" +
               "RO: None\n").format(3.)

    assert str(hm) == str_exp



import mogp_emulator
import numpy as np
from projectile import simulator, print_results

# simple MICE examples using the projectile demo

# Base design -- requires a list of parameter bounds if you would like to use
# uniform distributions. If you want to use different distributions, you
# can use any of the standard distributions available in scipy to create
# the appropriate ppf function (the inverse of the cumulative distribution).
# Internally, the code creates the design on the unit hypercube and then uses
# the distribution to map from [0,1] to the real parameter space.

lhd = mogp_emulator.LatinHypercubeDesign([(-5., 1.), (0., 1000.)])

###################################################################################

# first example -- run entire design internally within the MICE class.

# first argument is base design (required), second is simulator function (optional,
# but required if you want the code to run the simualtions internally)

# Other optional arguments include:
# n_samples (number of sequential design steps, optional, default is not specified
# meaning that you will specify when running the sequential design)
# n_init (size of initial design, default 10)
# n_cand (number of candidate points, default is 50)
# nugget (nugget parameter for design GP, default is to set adaptively)
# nugget_s (nugget parameter for candidate GP, default is 1.)

n_init = 5
n_samples = 20
n_cand = 100

md = mogp_emulator.MICEDesign(lhd, simulator, n_samples=n_samples, n_init=n_init, n_cand=n_cand)

md.run_sequential_design()

# get design and outputs

inputs = md.get_inputs()
targets = md.get_targets()

print("Example 1:")
print("Design inputs:\n", inputs)
print("Design targets:\n", targets)
print()

###################################################################################

# second example: run design manually

md2 = mogp_emulator.MICEDesign(lhd, n_init=n_init, n_cand=n_cand)

init_design = md2.generate_initial_design()

print("Example 2:")
print("Initial design:\n", init_design)

# run initial points manually

init_targets = np.array([simulator(s) for s in init_design])

# set initial targets

md2.set_initial_targets(init_targets)

# run 20 sequential design steps

for d in range(n_samples):
    next_point = md2.get_next_point()
    next_target = simulator(next_point)
    md2.set_next_target(next_target)

# look at design and outputs

inputs = md2.get_inputs()
targets = md2.get_targets()

print("Final inputs:\n", inputs)
print("Final targets:\n", targets)

# look at final GP emulator and make some predictions to compare with lhd

lhd_design = lhd.sample(n_init + n_samples)

gp_lhd = mogp_emulator.fit_GP_MAP(lhd_design, np.array([simulator(p) for p in lhd_design]))

gp_mice = mogp_emulator.GaussianProcess(inputs, targets)

gp_mice = mogp_emulator.fit_GP_MAP(inputs, targets)

test_points = lhd.sample(10)

print("LHD:")
print_results(test_points, gp_lhd(test_points))
print()
print("MICE:")
print_results(test_points, gp_mice(test_points))


import numpy as np
from projectile import simulator, print_results
import mogp_emulator

try:
    import matplotlib.pyplot as plt
    makeplots = True
except ImportError:
    makeplots = False

# An end-to-end tutorial illustrating model calibration using mogp_emulator

# First, we need to set up our experimental design. We would like our drag coefficient to be
# on a logarithmic scale and initial velocity to be on a linear scale. However, our simulator
# does the drag coefficient transformation for us, so we simply can specity the exponent on
# a linear scale.

# We will use a Latin Hypercube Design. To specify, we give the distribution that we would like
# the parameter to take. By default, we assume a uniform distribution between two endpoints, which
# we will use for this simulation.

# Once we construct the design, can draw a specified number of samples as shown.

lhd = mogp_emulator.LatinHypercubeDesign([(-5., 1.), (0., 1000.)])

n_simulations = 50
simulation_points = lhd.sample(n_simulations)
simulation_output = np.array([simulator(p) for p in simulation_points])

# Next, fit the surrogate GP model using MLE (MAP with uniform priors)
# Print out hyperparameter values as correlation lengths and sigma

gp = mogp_emulator.GaussianProcess(simulation_points, simulation_output)
gp = mogp_emulator.fit_GP_MAP(gp)

print("Correlation lengths = {}".format(np.sqrt(np.exp(-gp.theta[:2]))))
print("Sigma = {}".format(np.sqrt(np.exp(gp.theta[2]))))

# Validate emulator by comparing to true simulated value
# To compare with the emulator, use the predict method to get mean and variance
# values for the emulator predictions and see how many are within 2 standard
# deviations

n_valid = 10
validation_points = lhd.sample(n_valid)
validation_output = np.array([simulator(p) for p in validation_points])

predictions = gp.predict(validation_points)

print_results(validation_points, predictions.mean)

# Finally, perform history matching. Sample densely from the experimental design and
# determine which points are consistent with the data using the GP predictions
# We compute which points are "Not Ruled Out Yet" (NROY)

n_predict = 10000
prediction_points = lhd.sample(n_predict)

hm = mogp_emulator.HistoryMatching(gp=gp, coords=prediction_points, obs=[2000., 400.])

nroy_points = hm.get_NROY()

print("Ruled out {} of {} points".format(n_predict - len(nroy_points), n_predict))

# If plotting enabled, visualize results

if makeplots:
    plt.plot(prediction_points[nroy_points,0], prediction_points[nroy_points,1], "o", label="NROY points")
    plt.plot(simulation_points[:,0], simulation_points[:,1],"o", label="Simulation Points")
    plt.xlabel("log Drag Coefficient")
    plt.ylabel("Launch velocity (m/s)")
    plt.legend()
    plt.show()


import numpy as np
from scipy import linalg
from scipy.linalg import lapack

def check_cholesky_inputs(A):
    """
    Check inputs to cholesky routines

    This function is used by both specialized Cholesky routines to check inputs. It verifies
    that the input is a 2D square matrix and that all diagonal elements are positive (a
    necessary but not sufficient condition for positive definiteness). If these checks pass,
    it returns the input as a numpy array.

    :param A: The matrix to be inverted as an array of shape ``(n,n)``. Must be a symmetric positive
              definite matrix.
    :type A: ndarray or similar
    :returns: The matrix to be inverted as an array of shape ``(n,n)``. Must be a symmetric positive
              definite matrix.
    :rtype: ndarray
    """

    A = np.array(A)
    assert A.ndim == 2, "A must have shape (n,n)"
    assert A.shape[0] == A.shape[1], "A must have shape (n,n)"
    np.testing.assert_allclose(A.T, A)

    diagA = np.diag(A)
    if np.any(diagA <= 0.):
        raise linalg.LinAlgError("not pd: non-positive diagonal elements")

    return A

def jit_cholesky(A, maxtries = 5):
    """
    Performs Jittered Cholesky Decomposition

    Performs a Jittered Cholesky decomposition, adding noise to the diagonal of the matrix as needed
    in order to ensure that the matrix can be inverted. Adapted from code in GPy.

    On occasion, the matrix that needs to be inverted in fitting a GP is nearly singular. This arises
    when the training samples are very close to one another, and can be averted by adding a noise term
    to the diagonal of the matrix. This routine performs an exact Cholesky decomposition if it can
    be done, and if it cannot it successively adds noise to the diagonal (starting with 1.e-6 times
    the mean of the diagonal and incrementing by a factor of 10 each time) until the matrix can be
    decomposed or the algorithm reaches ``maxtries`` attempts. The routine returns the lower
    triangular matrix and the amount of noise necessary to stabilize the decomposition.

    :param A: The matrix to be inverted as an array of shape ``(n,n)``. Must be a symmetric positive
              definite matrix.
    :type A: ndarray
    :param maxtries: (optional) Maximum allowable number of attempts to stabilize the Cholesky
                     Decomposition. Must be a positive integer (default = 5)
    :type maxtries: int
    :returns: Lower-triangular factored matrix (shape ``(n,n)`` and the noise that was added to
              the diagonal to achieve that result.
    :rtype: tuple containing an ndarray and a float
    """

    A = check_cholesky_inputs(A)
    assert int(maxtries) > 0, "maxtries must be a positive integer"

    A = np.ascontiguousarray(A)
    L, info = lapack.dpotrf(A, lower = 1)
    if info == 0:
        return L, 0.
    else:
        diagA = np.diag(A)
        jitter = diagA.mean() * 1e-6
        num_tries = 1
        while num_tries <= maxtries and np.isfinite(jitter):
            try:
                L = linalg.cholesky(A + np.eye(A.shape[0]) * jitter, lower=True)
                return L, jitter
            except:
                jitter *= 10
            finally:
                num_tries += 1
        raise linalg.LinAlgError("not positive definite, even with jitter.")

    return L, jitter



"""This module provides classes and utilities for performing dimension
reduction.  Currently there is a single class :class:`mogp_emulator.gKDR` which implements
the method of Fukumizu and Leng [FL13]_.

Example: ::

  >>> from mogp_emulator import gKDR
  >>> import numpy as np
  >>> X = np.array([[0.0, 0.0], [0.0, 1.0], [1.0, 0.0], [1.0, 1.0]])
  >>> Y = np.array([0.0, 1.0, 5.0, 6.0])
  >>> xnew = np.array([0.5, 0.5])
  >>> dr = gKDR(X, Y, 1)
  >>> dr(xnew)
  array([0.60092477])

In this example, the reduction was performed from a two- to a
one-dimensional input space.  The value returned by ``dr(xnew)`` is
the input coordinate `xnew` transformed to the reduced space.

The following example illustrates how to perform Gaussian process
regression on the reduced input space:

::

  >>> import numpy as np
  >>> from mogp_emulator import gKDR, GaussianProcess, fit_GP_MAP

  ### generate some training data (from the function f)

  >>> def f(x):
  ...     return np.sqrt(x[0] + np.sin(0.1 * x[1]))

  >>> X = np.mgrid[0:10,0:10].T.reshape(-1,2)/10.0
  >>> print(X)
  [[0.  0. ]
   [0.1 0. ]
   [0.2 0. ]
   [0.3 0. ]
   [0.4 0. ]
   [0.5 0. ]
   [0.6 0. ]
   [0.7 0. ]
   [0.8 0. ]
   [0.9 0. ]
   [0.  0.1]
   [0.1 0.1]
   ...
   [0.8 0.9]
   [0.9 0.9]]

  >>> Y = np.apply_along_axis(f, 1, X)

  ### reduced input space
  >>> dr = gKDR(X, Y, K=1)

  ### train a Gaussian Process with reduced inputs
  >>> gp = GaussianProcess(dr(X), Y)
  >>> gp = fit_GP_MAP(gp)

  ### make a prediction (given an input in the reduced space)
  >>> Xnew = np.array([0.12, 0.37])
  >>> gp.predict(dr(Xnew))
  (array([0.398083]), ...)

  >>> f(Xnew)
  0.396221

"""

import sys
import numpy as np
from scipy.spatial.distance import cdist, pdist, squareform
from mogp_emulator.utils import k_fold_cross_validation, integer_bisect

def gram_matrix(X, k):
    """Computes the Gram matrix of `X`

    :type X: ndarray

    :param X: Two-dimensional numpy array, where rows are feature
              vectors

    :param k: The covariance function

    :returns: The gram matrix of `X` under the kernel `k`, that is,
              :math:`G_{ij} = k(X_i, X_j)`
    """
    ## note: do not use squareform(pdist(X, k)) here, since it assumes
    ## that dist(x,x) == 0, which might not be the case for an arbitrary k.
    return cdist(X, X, k)

def gram_matrix_sqexp(X, sigma2):
    r"""Computes the Gram matrix of `X` under the squared expontial kernel.
    Equivalent to, but more efficient than, calling ``gram_matrix(X,
    k_sqexp)``

    :type X: ndarray

    :param X: Two-dimensional numpy array, where rows are feature
              vectors

    :param sigma2: The variance parameter of the squared exponential kernel

    :returns: The gram matrix of `X` under the squared exponential
              kernel `k_sqexp` with variance parameter `sigma2` (:math:`=\sigma^2`), that
              is, :math:`G_{ij} = k_{sqexp}(X_i, X_j; \sigma^2)`

    """
    return np.exp(-0.5 * squareform(pdist(X, 'sqeuclidean')) / sigma2)


def median_dist(X):
    """Return the median of the pairwise (Euclidean) distances between
    each row of X
    """
    return np.median(pdist(X))


class gKDR(object):

    """Dimension reduction by the gKDR method.

    See link [Fukumizu1]_ (and in particular, [FL13]_) for details of
    the method.

    Note that this is a simpler and faster method than the original
    "KDR" method by the same authors (but with an added
    approximation).  The KDR method will be implemented separately.

    An instance of this class is callable, with the ``__call__``
    method taking an input coordinate and mapping it to a reduced
    coordinate.

    Note that this class currently implements a *direct* translation
    of the Matlab implementation of KernelDeriv (see link above) into
    Python/NumPy.  It is due to be replaced with a Fortran
    implementation, but this should not affect the interface.
    """

    def __init__(self, X, Y, K=None, X_scale = 1.0, Y_scale = 1.0, EPS=1E-8, SGX=None, SGY=None):
        """Create a gKDR object

        Given some `M`-dimensional inputs (explanatory variables) `X`,
        and corresponding one-dimensional outputs (responses) `Y`, use
        the gKDR method to produce a reduced version of the input
        space with `K` dimensions.

        :type X: ndarray, of shape (N, M)
        :param X: `N` rows of `M` dimensional input vectors

        :type Y: ndarray, of shape (N,)
        :param Y: `N` response values

        :type K: integer
        :param K: The number of reduced dimensions to use (`0 <= K <= M`).

        :type EPS: float
        :param EPS: The regularization parameter, default `1e-08`; `EPS >= 0`

        :type X_scale: float
        :param X_scale: Optional, default `1.0`.  If SGX is None (the default), scale the
                        automatically determined value for SGX by X_scale.  Otherwise ignored.

        :type Y_scale: float
        :param Y_scale: Optional, default `1.0`.  If SGY is None (the default), scale the
                        automatically determined value for SGY by Y_scale.  Otherwise ignored.

        :type SGX: float | NoneType
        :param SGX: Optional, default `None`. The kernel parameter representing the
                    scale of variation on the input space.  If `None`, then the median distance
                    between pairs of input points (`X`) is used (as computed by
                    :func:`mogp_emulator.DimensionReduction.median_dist`).  If a float is
                    passed, then this must be positive.

        :type SGY: float | NoneType
        :param SGY: Optional, default `None`. The kernel parameter representing the
                    scale of variation on the output space.  If `None`, then the median distance
                    between pairs of output values (`Y`) is used (as computed by
                    :func:`mogp_emulator.DimensionReduction.median_dist`).  If a float is
                    passed, then this must be positive.
        """

        ## Note: see the Matlab implementation ...

        N, M = np.shape(X)

        ## default K: use the entire input space
        if K is None:
            K = M

        assert(K >= 0 and K <= M)
        assert(EPS >= 0)
        assert(SGX is None or SGX > 0.0)
        assert(SGY is None or SGY > 0.0)

        Y = np.reshape(Y, (N,1))

        if SGX is None:
            SGX = X_scale * median_dist(X)
        if SGY is None:
            SGY = Y_scale * median_dist(Y)

        I = np.eye(N)

        SGX2 = max(SGX*SGX, sys.float_info.min)
        SGY2 = max(SGY*SGY, sys.float_info.min)

        Kx = gram_matrix_sqexp(X, SGX2)
        Ky = gram_matrix_sqexp(Y, SGY2)

        Dx = np.reshape(np.tile(X,(N,1)), (N,N,M), order='F').copy()
        Xij = Dx - np.transpose(Dx, (1,0,2))
        Xij = Xij / SGX2
        H = Xij * np.tile(Kx[:,:,np.newaxis], (1,1,M))

        tmp = np.linalg.solve(Kx + N*EPS*I, Ky)
        F = np.linalg.solve((Kx + N*EPS*I).T, tmp.T).T

        Hm = np.reshape(H,(N,N*M), order='F')
        HH = np.reshape(Hm.T @ Hm, (N,M,N,M), order='F')
        HHm = np.reshape(np.transpose(HH, (0,2,1,3)), (N*N,M,M), order='F')
        Fm = np.tile(np.reshape(F, (N*N,1,1), order='F'), (1,M,M))
        R = np.reshape(np.sum(HHm * Fm, 0), (M,M), order='F')

        L, V = np.linalg.eigh(R)

        assert(np.allclose(V.imag, 0.0))

        idx = np.argsort(L, 0)[::-1] # sort descending

        # record B, along with some bookkeeping parameters
        self.X_scale = X_scale
        self.Y_scale = Y_scale
        self.K = K
        self.B = V[:, idx]


    def __call__(self, X):
        """Calling a gKDR object with a vector of N inputs returns the inputs
        mapped to the reduced space.

        :type X: ndarray, of shape `(N, M)`
        :param X: `N` coordinates (rows) in the unreduced `M`-dimensional space

        :rtype: ndarray, of shape `(N, K)`
        :returns:  `N` coordinates (rows) in the reduced `K`-dimensional space
        """
        return X @ self.B[:,0:self.K]


    @staticmethod
    def _compute_loss( X, Y, train_model, cross_validation_folds, *params, **kwparams):
        """Compute the L1 loss of a model (produced by calling train_model), via
        cross validation.  The model is trained on input parameters `x` that
        are first reduced via the dimension reduction procedure produced by
        calling ``gKDR(x, y, *params)``.

        :type X: ndarray, of shape (N, M)
        :param X: `N` input points with dimension `M`

        :type Y: ndarray, of shape (N,)
        :param Y: the `N` model observations, corresponding to each

        :type train_model: callable with the signature
                           `(ndarray, ndarray) -> ndarray -> ndarray`
        :param train_model: a callable, that when called with model inputs X
                            (shape `(Ntrain, M)`) and Y (shape `(Ntrain, M)`),
                            returns a "model", which is another callable, taking
                            an array (shape `(Npredict, M)`) of the points where
                            a prediction is desired, and returning an array
                            (shape `(Npredict,)`) of the corresponding
                            predictions.

        :type cross_validation_folds: integer
        :param cross_validation_folds: Use this many folds for cross-validation
                                       when tuning the parameters.

        :type params: tuple
        :param params: parameters to pass to :meth:`mogp_emulator.gKDR.__init__`

        :type kwparams: dict
        :param kwparams: keyword parameters to pass to
                         :meth:`mogp_emulator.gKDR.__init__`
        """

        ## combine input and output arrays, such that if
        ## X[i] = [1,2,3] and Y[i] = 4, then XY[i] = [1,2,3,4].
        ## That is,
        ## XY[:, -1] == Y
        ## XY[:, 0:-1] == X
        ##
        XY = np.hstack((X, Y[:,np.newaxis]))

        err = []
        for train, validate in k_fold_cross_validation(XY, cross_validation_folds):
            train = np.array(train)
            validate = np.array(validate)

            dr = gKDR(train[:,0:-1], train[:,-1], *params, **kwparams)

            model = train_model(dr(train[:,0:-1]), train[:,-1])

            error_L1 = np.mean(np.abs(validate[:,-1] - model(dr(validate[:,0:-1]))))
            err.append(error_L1)

        return np.mean(err)


    @classmethod
    def tune_parameters(cls, X, Y, train_model, cXs=None, cYs=None,
                        maxK=None, cross_validation_folds=5,
                        verbose=False):
        """Constructs a gKDR model with the structural dimension (K) and
        kernel scale parameters (cX, cY) that approximately minimize
        the L1 error between Y and the trained model (resulting from
        calling `train_model` on X and Y).

        Currently, this works as follows.  For each choice of cX and
        cY in `cXs` and `cYs`, find K by starting from a K of 1 and
        doubling K until the loss increases (using the value of K just
        before), or until K equals the input dimension (or maxK if
        specified).  The resulting choice of `(cX, cY, K)` is then
        taken as the minimum such choice over the cX, cY.

        :type X: ndarray, of shape (N, M)
        :param X: `N` input points with dimension `M`

        :type Y: ndarray, of shape (N,)
        :param Y: the `N` model observations, corresponding to each

        :type train_model: callable with the signature
                           `(ndarray, ndarray) -> ndarray -> ndarray`
        :param train_model: a callable, that when called with model inputs X
                            (shape `(Ntrain, M)`) and Y (shape `(Ntrain, M)`),
                            returns a "model", which is another callable, taking
                            an array (shape `(Npredict, M)`) of the points where
                            a prediction is desired, and returning an array
                            (shape `(Npredict,)`) of the corresponding
                            predictions.

        :type cXs: Iterable of `float`, or `NoneType`
        :param cXs: (optional, default None). The scale parameter for `X` in the
                    dimension reduction kernel.  Passed as the parameter
                    `X_scale` to the gKDR constructor
                    (:meth:`mogp_emulator.gKDR.__init__`). If None, `[0.5, 1,
                    5.0]` is used.

        :type cYs: Iterable of `float`, or `NoneType`
        :param cYs: (optional, default None). The scale parameter for `Y` in the
                    dimension reduction kernel.  Passed as the parameter
                    `Y_scale` to the gKDR constructor
                    (:meth:`mogp_emulator.gKDR.__init__`). If None, `[0.5, 1,
                    5.0]` is used.

        :type maxK: integer, or NoneType
        :param maxK: (optional default `None`). The largest structural dimension
                     to consider in the optimization.  This is useful when there
                     is a known bound on the dimension, to stop e.g. poor values
                     of cX or cY needlessly extending the search.  It is a good
                     idea to choose this parameter generously.

        :type cross_validation_folds: integer
        :param cross_validation_folds: (optional, default is 5): Use this many
                                       folds for cross-validation when tuning
                                       the parameters.

        :type verbose: bool
        :param verbose: produce a log to stdout of the optimization?

        :returns: A pair of: the gKDR object with parameters tuned according to
                  the above method, and a number representing the L1 loss of the
                  model trained on inputs as reduced by this dimension reduction
                  object.  :rtype: pair of a gKDR and a non-negative float

        *Example*

        Tune the structural dimension and lengthscale parameters
        within the kernel, minimizing the the loss from a
        Gaussian process regression:

          >>> from mogp_emulator import gKDR
          >>> from mogp_emulator import GaussianProcess
          >>> X = ...
          >>> Y = ...
          >>> dr, loss = gKDR.tune_parameters(X, Y, GaussianProcess.train_model)
          >>> gp = GaussianProcess(dr(X), Y)

        Or, specifying some optional parameters for the lengthscales,
        the maximum value of `K` to use, the number of folds for
        cross-validation, and producing verbose output:

          >>> dr, loss = gKDR.tune_parameters(X, Y, GaussianProcess.train_model,
          ...                                 cXs = [0.5, 1.0, 2.0], cYs = [2.0],
          ...                                 maxK = 25, cross_validation_folds=4, verbose = True)

        """

        N, M = np.shape(X)

        if cXs is None:
            cXs = [0.5, 1.0, 5.0]

        if cYs is None:
            cYs = [0.5, 1.0, 5.0]

        if maxK is None:
            maxK = M

        assert(maxK >= 1 and maxK <= M)

        def compute_loss(*params):
            return gKDR._compute_loss(
                X, Y, train_model, cross_validation_folds, *params)

        # Search for K and scale parameters that together
        # (approximately) minimize the loss, which could be quite
        # noisy.  For each choice of the scale parameters (cX, cY), K
        # starts at 1 and doubles until the loss increases, or until K
        # equals the input dimension.  Keep the minimum over all of
        # the trial parameter sets.
        #
        min_loss = np.inf
        argmin_loss = None
        for cX in cXs:
            for cY in cYs:
                loss = np.inf
                params = None
                k = 1
                while (k <= maxK):
                    old_params, params = params, (k, cX, cY)
                    old_loss, loss = loss, compute_loss(*params)

                    if verbose:
                        print("loss(K={}, X_scale={}, Y_scale={}) = {}"\
                              .format(*params, loss))

                    if old_loss < loss:
                        if old_loss < min_loss:
                            min_loss = old_loss
                            argmin_loss = old_params
                        break
                    elif k == maxK:
                        if loss < min_loss:
                            min_loss = loss
                            argmin_loss = params
                        break
                    elif 2*k > maxK:
                        k = maxK
                    else:
                        k *= 2

        # construct the object with the full set of data, and
        # parameters found above
        #
        dr = gKDR(X, Y, *argmin_loss)

        return (dr, min_loss)



import numpy as np
import scipy.stats
from scipy.linalg import LinAlgError
from scipy.optimize import minimize
from mogp_emulator.GaussianProcess import GaussianProcess
from mogp_emulator.MultiOutputGP import MultiOutputGP
from multiprocessing import Pool
from functools import partial
import platform

def fit_GP_MAP(*args, n_tries=15, theta0=None, method="L-BFGS-B", **kwargs):
    """
    Fit one or more Gaussian Processes by attempting to minimize the negative log-posterior

    Fits the hyperparameters of one or more Gaussian Processes by attempting to minimize
    the negative log-posterior multiple times from a given starting location and using
    a particular minimization method. The best result found among all of the attempts is
    returned, unless all attempts to fit the parameters result in an error (see below).

    The arguments to the method can either be an existing ``GaussianProcess`` or
    ``MultiOutputGP`` instance, or a list of arguments to be passed to the ``__init__``
    method of ``GaussianProcess`` or ``MultiOutputGP`` if more than one output is detected.
    Keyword arguments for creating a new ``GaussianProcess`` or ``MultiOutputGP`` object can
    either be passed as part of the ``*args`` list or as keywords (if present in ``**kwargs``, they
    will be extracted and passed separately to the ``__init__`` method).

    If the method encounters an overflow (this can result because the parameter values stored are
    the logarithm of the actual hyperparameters to enforce positivity) or a linear algebra error
    (occurs when the covariance matrix cannot be inverted, even with additional noise added along
    the diagonal if adaptive noise was selected), the iteration is skipped. If all attempts to find
    optimal hyperparameters result in an error, then the method raises an exception.

    The ``theta0`` parameter is the point at which the first iteration will start. If more than
    one attempt is made, subsequent attempts will use random starting points. If you are fitting
    Multiple Outputs, then this argument can take any of the following forms: (1) None (random
    start points for all emulators), (2) a list of numpy arrays or ``NoneTypes`` with length
    ``n_emulators``, (3) a numpy array of shape ``(n_params,)`` or ``(n_emulators, n_params)``
    which with either use the same start point for all emulators or the specified start
    point for all emulators. Note that if you us a numpy array, all emulators must have the
    same number of parameters, while using a list allows more flexibility.

    The user can specify the details of the minimization method, using any of the gradient-based
    optimizers available in ``scipy.optimize.minimize``. Any additional parameters beyond the method
    specification can be passed as keyword arguments.

    The function returns a fit ``GaussianProcess`` or ``MultiOutputGP`` instance, either the original
    one passed to the function, or the new one created from the included arguments.

    :param ``*args``: Either a single ``GaussianProcess`` or ``MultiOutputGP`` instance,
                      or arguments to be passed to the ``__init__`` method when creating a new
                      ``GaussianProcess`` or ``MultiOutputGP`` instance.
    :param n_tries: Number of attempts to minimize the negative log-posterior function.
                    Must be a positive integer (optional, default is 15)
    :type n_tries: int
    :param theta0: Initial starting point for the first iteration. If present, must be
                   array-like with shape ``(n_params,)`` based on the specific
                   ``GaussianProcess`` being fit. If a ``MultiOutputGP`` is being fit
                   it must be a list of length ``n_emulators`` with each entry as either
                   ``None`` or a numpy array of shape ``(n_params,)``, or a numpy array
                   with shape ``(n_emulators, n_params)`` (note that if the various emulators
                   have different numbers of parameters, the numpy array option will not work).
                   If ``None`` is given, then a random value is chosen. (Default is ``None``)
    :type theta0: None or ndarray
    :param method: Minimization method to be used. Can be any gradient-based optimization
                   method available in ``scipy.optimize.minimize``. (Default is ``'L-BFGS-B'``)
    :type method: str
    :param ``**kwargs``: Additional keyword arguments to be passed to ``GaussianProcess.__init__``,
                         ``MultiOutputGP.__init__``, or the minimization routine. Relevant parameters
                         for the GP classes are automatically split out from those used in the
                         minimization function. See available parameters in the corresponding functions
                         for details.
    :returns: Fit GP or Multi-Output GP instance
    :rtype: GaussianProcess or MultiOutputGP
    """

    if len(args) == 1:
        gp = args[0]
        if isinstance(gp, MultiOutputGP):
            return _fit_MOGP_MAP(gp, n_tries, theta0, method, **kwargs)
        elif isinstance(gp, GaussianProcess):
            return _fit_single_GP_MAP(gp, n_tries, theta0, method, **kwargs)
        else:
            raise TypeError("single arg to fit_GP_MAP must be a GaussianProcess or MultiOutputGP instance")
    elif len(args) < 2:
        raise TypeError("missing required inputs/targets arrays to GaussianProcess")
    else:
        gp_kwargs = {}
        for key in ["mean", "kernel", "priors", "nugget", "inputdict", "use_patsy"]:
            if key in kwargs:
                gp_kwargs[key] = kwargs[key]
                del kwargs[key]
        try:
            gp = GaussianProcess(*args, **gp_kwargs)
            return _fit_single_GP_MAP(gp, n_tries, theta0, method, **kwargs)
        except AssertionError:
            gp = MultiOutputGP(*args, **gp_kwargs)
            return _fit_MOGP_MAP(gp, n_tries, theta0, method, **kwargs)

def _fit_single_GP_MAP(gp, n_tries=15, theta0=None, method='L-BFGS-B', **kwargs):
    """
    Fit hyperparameters using MAP for a single GP

    Returns a single GP object that has its hyperparameters fit to the MAP value.
    Accepts keyword arguments passed to scipy's minimization routine.
    """

    assert isinstance(gp, GaussianProcess)

    n_tries = int(n_tries)
    assert n_tries > 0, "number of attempts must be positive"

    np.seterr(divide = 'raise', over = 'raise', invalid = 'raise')

    logpost_values = []
    theta_values = []

    theta_startvals = 5.*(np.random.rand(n_tries, gp.n_params) - 0.5)
    if not theta0 is None:
        theta0 = np.array(theta0)
        assert theta0.shape == (gp.n_params,), "theta0 must be a 1D array with length n_params"
        theta_startvals[0,:] = theta0

    for theta in theta_startvals:
        try:
            min_dict = minimize(gp.logposterior, theta, method = method,
                                jac = gp.logpost_deriv, options = kwargs)

            min_theta = min_dict['x']
            min_logpost = min_dict['fun']

            logpost_values.append(min_logpost)
            theta_values.append(min_theta)
        except LinAlgError:
            print("Matrix not positive definite, skipping this iteration")
        except FloatingPointError:
            print("Floating point error in optimization routine, skipping this iteration")

    if len(logpost_values) == 0:
        raise RuntimeError("Minimization routine failed to return a value")

    logpost_values = np.array(logpost_values)
    idx = np.argmin(logpost_values)

    gp.fit(theta_values[idx])

    return gp

def _fit_single_GP_MAP_bound(gp, theta0, n_tries, method, **kwargs):
    "fitting function accepting theta0 as an argument for parallelization"

    return _fit_single_GP_MAP(gp, n_tries=n_tries, theta0=theta0, method=method, **kwargs)

def _fit_MOGP_MAP(gp, n_tries=15, theta0=None, method='L-BFGS-B', **kwargs):
    """
    Fit hyperparameters using MAP for multiple GPs in parallel

    Uses Python Multiprocessing to fit GPs in parallel by calling the above routine for a single
    GP for each of the emulators in the MOGP class. Returns a MultiOutputGP object where all
    emulators have been fit to the MAP value.

    Accepts a ``processes`` argument (integer or None) as a keyword to control the number of
    subprocesses used to fit the individual GPs in parallel. Must be positive. Default is ``None``.
    """

    assert isinstance(gp, MultiOutputGP)

    try:
        processes = kwargs['processes']
        del kwargs['processes']
    except KeyError:
        processes = None

    assert int(n_tries) > 0, "n_tries must be a positive integer"

    if theta0 is None:
        theta0 = [ None ]*gp.n_emulators
    else:
        if isinstance(theta0, np.ndarray):
            if theta0.ndim == 1:
                theta0 = [theta0]*gp.n_emulators
            else:
                assert theta0.ndim == 2, "theta0 must be a 1D or 2D array"
                assert theta0.shape[0] == gp.n_emulators, "bad shape for fitting starting points"
        elif isinstance(theta0, list):
            assert len(theta0) == gp.n_emulators, "theta0 must be a list of length n_emulators"

    if not processes is None:
        processes = int(processes)
        assert processes > 0, "number of processes must be positive"

    n_tries = int(n_tries)

    # partial(fit_GP_MAP, )

    if platform.system() == "Windows":
        fit_MOGP = [fit_GP_MAP(emulator, n_tries=n_tries, theta0=t0, method=method, **kwargs)
                    for (emulator, t0) in zip(gp.emulators, theta0)]
    else:
        with Pool(processes) as p:
            fit_MOGP = p.starmap(partial(_fit_single_GP_MAP_bound, n_tries=n_tries, method=method, **kwargs),
                                 [(emulator, t0) for (emulator, t0) in zip(gp.emulators, theta0)])

    gp.emulators = fit_MOGP

    return gp


import numpy as np
import pytest
from numpy.testing import assert_allclose
from ..MCMC import MH_proposal, MCMC_step, sample_MCMC, autothin_samples

def test_MH_proposal():
    "test the Metropolis-Hastings proposal distribution"
    
    np.random.seed(438)
    
    current_param = np.zeros(2)
    cov = np.eye(2)
    
    assert_allclose(MH_proposal(current_param, cov), np.array([-0.392811129690724 ,  0.4478392215027822]))
    
    with pytest.raises(AssertionError):
        MH_proposal(np.zeros((2,2)), cov)
        
    with pytest.raises(AssertionError):
        MH_proposal(np.zeros(2), np.ones((2, 2, 2)))
        
    with pytest.raises(AssertionError):
        MH_proposal(np.zeros(2), np.ones((2, 3)))
        
    with pytest.raises(AssertionError):
        MH_proposal(np.zeros(3), np.ones((2, 2)))
        
    with pytest.raises(AssertionError):
        MH_proposal(np.zeros(2), -1.*np.eye(2))

def test_MCMC_step():
    "test the MCMC step routine"
    
    def loglikelihood(x):
        return 1.
        
    current_param = np.zeros(2)
    cov = np.eye(2)
    
    next_point_expected = np.array([-0.392811129690724 ,  0.4478392215027822])
    
    np.random.seed(438)
    
    next_point, accept = MCMC_step(loglikelihood, current_param, cov)
    
    assert_allclose(next_point, next_point_expected)
    assert accept
    
    def loglikelihood(x):
        return 100.*np.sum(x)

    np.random.seed(438)
    
    next_point, accept = MCMC_step(loglikelihood, current_param, cov, 1.)
    
    assert_allclose(next_point, next_point_expected)
    assert accept
    
    np.random.seed(438)
    
    next_point, accept = MCMC_step(loglikelihood, current_param, cov, -1.)
    
    assert_allclose(next_point, next_point_expected)
    assert not accept
    
    with pytest.raises(AssertionError):
        MCMC_step(loglikelihood, current_param, cov, 0.)
        
    def loglikelihood(x, y):
        return -1.
    
    with pytest.raises(AssertionError):
        MCMC_step(loglikelihood, current_param, cov)
        
    with pytest.raises(AssertionError):
        MCMC_step(2., current_param, cov)
        
def test_sample_MCMC():
    "test MCMC sampling"
    
    def loglikelihood(x):
        return 100.*np.sum(x)
    
    current_param = np.zeros(2)
    cov = np.eye(2)
    
    samples_expected = np.array([[0., 0.],
                                [-0.392811129690724 ,  0.4478392215027822],
                                [-1.6681438506139665,  2.697940030289671 ],
                                [-1.6681438506139665,  2.697940030289671 ]])
    rejected_expected = np.array([[-3.4166078360234584,  2.5824123019889043]])
    acceptance_expected = 0.75
    first_lag_expected = np.array([0.2886356417500504, 0.2824351401835904])
    
    np.random.seed(438)
    
    with pytest.warns(Warning):
        samples, rejected, acceptance, first_lag = sample_MCMC(loglikelihood, current_param, cov, n_samples = 4, thin = 0)
    
    assert_allclose(samples, samples_expected)
    assert_allclose(rejected, rejected_expected)
    assert_allclose(acceptance, acceptance_expected)
    assert_allclose(first_lag, first_lag_expected)
    
    np.random.seed(438)
    
    samples, rejected, acceptance, first_lag = sample_MCMC(loglikelihood, current_param, cov, n_samples = 4, thin = 1)
    
    assert_allclose(samples, samples_expected)
    assert_allclose(rejected, rejected_expected)
    assert_allclose(acceptance, acceptance_expected)
    assert_allclose(first_lag, first_lag_expected)
    
    samples_expected = np.array([[0., 0.],
                                [-1.6681438506139665,  2.697940030289671 ]])
    rejected_expected = np.array([[-3.4166078360234584,  2.5824123019889043]])
    acceptance_expected = 0.75
    first_lag_expected = np.array([-0.5, -0.5])
    
    np.random.seed(438)
    
    samples, rejected, acceptance, first_lag = sample_MCMC(loglikelihood, current_param, cov, n_samples = 4, thin = 2)
    
    assert_allclose(samples, samples_expected)
    assert_allclose(rejected, rejected_expected)
    assert_allclose(acceptance, acceptance_expected)
    assert_allclose(first_lag, first_lag_expected)
    
    with pytest.raises(AssertionError):
        sample_MCMC(loglikelihood, current_param, cov, n_samples = -1)
        
    with pytest.raises(AssertionError):
        sample_MCMC(loglikelihood, current_param, cov, 1000, -1)
        
    with pytest.raises(AssertionError):
        sample_MCMC(loglikelihood, np.zeros((2,3)), cov, 1000, -1)

def test_autothin_samples():
    "test the autothinning routine"
    
    a = np.zeros(100)
    a[45:55] = 1.
    
    assert autothin_samples(a) == 7
    
    a = np.zeros((100, 2))
    a[45:55, 0] = 1.
    a[40:60, 1] = 1.
    
    assert autothin_samples(a) == 11
    
    a = np.zeros(9)
    
    with pytest.warns(Warning):
        assert autothin_samples(a) == 1
    
    a = np.zeros((2, 2, 2))
    
    with pytest.raises(AssertionError):
        autothin_samples(a)
    




import numpy as np
from scipy.special import gamma

class Prior(object):
    """
    Generic Prior Object
    """
    def logp(self, x):
        """
        Computes log probability at a given value
        """
        raise NotImplementedError

    def dlogpdtheta(self, x):
        """
        Computes derivative of log probability at a given value
        """
        raise NotImplementedError

    def d2logpdtheta2(self, x):
        """
        Computes second derivative of log probability at a given value
        """
        raise NotImplementedError

class NormalPrior(Prior):
    """
    Normal Distribution Prior object

    Admits input values from -inf/+inf, and no transformations are assumed. Thus, for mean function
    hyperparameters this produces a normal distribution with given mean and variance, and for
    covariance/nugget hyperparameters this produces a lognormal distribution with given log mean and
    variance.

    Take two parameters: mean and std. mean can take any numeric value, while std must be positive.
    """
    def __init__(self, mean, std):
        self.mean = mean
        assert std > 0., "std parameter must be positive"
        self.std = std

    def logp(self, x):
        """
        Computes log probability at a given value
        """
        return -0.5*((x - self.mean)/self.std)**2 - np.log(self.std) - 0.5*np.log(2.*np.pi)

    def dlogpdtheta(self, x):
        """
        Computes derivative of log probability at a given value
        """
        return -(x - self.mean)/self.std**2

    def d2logpdtheta2(self, x):
        """
        Computes second derivative of log probability at a given value
        """
        return -self.std**(-2)

class GammaPrior(Prior):
    r"""
    Gamma Distribution Prior object

    Admits input values from -inf/+inf assumed on a logarithmic scale, and transforms by taking the
    exponential of the input. Thus, this is assumed to be appropriate for covariance hyperparameters
    where such transformations are assumed when computing the covariance function.

    Take two parameters: shape :math:`{\alpha}` and scale :math:`{\beta}`. Both must be positive,
    and they are defined such that

    :math:`{p(x) = \frac{\beta^{-\alpha}x^{\beta - 1}}{\Gamma(/alpha)} \exp(-x/\beta)}`
    """
    def __init__(self, shape, scale):
        assert shape > 0., "shape parameter must be positive"
        self.shape = shape
        assert scale > 0., "scale parameter must be positive"
        self.scale = scale

    def logp(self, x):
        """
        Computes log probability at a given value
        """
        return (-self.shape*np.log(self.scale) - np.log(gamma(self.shape)) +
                (self.shape - 1.)*x - np.exp(x)/self.scale)

    def dlogpdtheta(self, x):
        """
        Computes derivative of log probability at a given value
        """
        return (self.shape - 1.) - np.exp(x)/self.scale

    def d2logpdtheta2(self, x):
        """
        Computes second derivative of log probability at a given value
        """
        return -np.exp(x)/self.scale

class InvGammaPrior(Prior):
    r"""
    Inverse Gamma Distribution Prior object

    Admits input values from -inf/+inf assumed on a logarithmic scale, and transforms by taking the
    exponential of the input. Thus, this is assumed to be appropriate for covariance hyperparameters
    where such transformations are assumed when computing the covariance function.

    Take two parameters: shape :math:`{\alpha}` and scale :math:`{\beta}`. Both must be positive,
    and they are defined such that

    :math:`{p(x) = \frac{\beta^{\alpha}x^{-\beta - 1}}{\Gamma(/alpha)} \exp(-\beta/x)}`
    """
    def __init__(self, shape, scale):
        assert shape > 0., "shape parameter must be positive"
        self.shape = shape
        assert scale > 0., "scale parameter must be positive"
        self.scale = scale

    def logp(self, x):
        """
        Computes log probability at a given value
        """
        return (self.shape*np.log(self.scale) - np.log(gamma(self.shape)) -
                (self.shape + 1.)*x - self.scale*np.exp(-x))

    def dlogpdtheta(self, x):
        """
        Computes derivative of log probability at a given value
        """
        return -(self.shape + 1.) + self.scale*np.exp(-x)

    def d2logpdtheta2(self, x):
        """
        Computes second derivative of log probability at a given value
        """
        return -self.scale*np.exp(-x)
